import enum
"""
Wrapper module for JSON ARM model
"""

# Operations on ARM MODEL
CONCRETE_DOMAINS = 'concreteDomains'
ABSTRACT_DOMAINS = 'abstractDomains'
RELATION_TYPES = 'relationTypes'
ANY_TYPE = 'anyType'
SELF = 'self'
OID = 'OID'
NOTES = 'notes'

def domains(arm_model):
    assert(CONCRETE_DOMAINS in arm_model and ABSTRACT_DOMAINS in arm_model)
    return arm_model[CONCRETE_DOMAINS]+arm_model[ABSTRACT_DOMAINS]


def get_relation_type(arm_model, relation_name):
    assert(RELATION_TYPES in arm_model)
    for relation in get_relation_types(arm_model):
        if relation_name == relation[RELATION_NAME]:
            return relation
    return None


def get_relation_types(arm_model):
    assert(RELATION_TYPES in arm_model)
    return arm_model[RELATION_TYPES]


def has_notes(arm_model):
    return NOTES in arm_model

def get_notes(arm_model):
    return arm_model[NOTES]

def create_arm_model():
    return { CONCRETE_DOMAINS: list(), ABSTRACT_DOMAINS: list(), RELATION_TYPES: list() }


class DomainType(enum.Enum):
    ABSTRACT = enum.auto()
    CONCRETE = enum.auto()
    UNKNOWN = enum.auto()


def add_domain(arm_model, domain, domain_type):
    assert(CONCRETE_DOMAINS in arm_model and ABSTRACT_DOMAINS in arm_model)
    if domain_type==DomainType.ABSTRACT:
        arm_model[ABSTRACT_DOMAINS].append(domain)
    else:
        arm_model[CONCRETE_DOMAINS].append(domain)


def add_relation_type(arm_model, name):
    relation_type = create_relation(name)
    arm_model[RELATION_TYPES].append(relation_type)
    return relation_type

def add_notes(arm_model, notes):
    arm_model[NOTES] = notes

# Operations on ARM Relation
RELATION_NAME = 'name'
RELATION_ATTRIBUTES = 'attributes'
RELATION_CONSTRAINTS = 'constraints'
ATTRIBUTE_NAME = 'name'
ATTRIBUTE_DATA_TYPE = 'dataType'


def constraints(relation_type):
    if RELATION_CONSTRAINTS in relation_type:
        return relation_type[RELATION_CONSTRAINTS] # Maybe duplicates
    else:
        return None


def relation_type_name(relation_type):
    assert(RELATION_NAME in relation_type)
    return relation_type[RELATION_NAME]


def attribute_names(relation_type):
    assert(RELATION_ATTRIBUTES in relation_type)
    return { attribute_name for attribute_name in relation_attributes(relation_type)}


def relation_attributes(relation):
    assert(RELATION_ATTRIBUTES in relation)
    attributes = dict()
    for attribute in relation[RELATION_ATTRIBUTES]:
        attributes[attribute[ATTRIBUTE_NAME]] = attribute[ATTRIBUTE_DATA_TYPE]
    return attributes

def is_attribute(relation, name):
    return name in attribute_names(relation)


def get_data_type(relation, attribute_name):
    return relation_attributes(relation)[attribute_name]


def split_attributes(relation):
    all_attributes = relation_attributes(relation).copy()
    fk_result = dict()
    for constraint in foreign_key_constraints(constraints(relation)):
        attribute_name = foreign_key_attribute(constraint)
        fk_result[attribute_name] = all_attributes[attribute_name]
        del all_attributes[attribute_name]
    non_fk_result = all_attributes
    return fk_result, non_fk_result

def foreign_key_attributes(relation):
    return split_attributes(relation)[0]

def non_fk_attributes(relation):
    return split_attributes(relation)[1]

def get_foreign_key_constraint(relation, foreign_key_attribute_name):
    for constraint in foreign_key_constraints(constraints(relation)):
        if foreign_key_attribute_name ==  foreign_key_attribute(constraint):
            return constraint
    return None


def is_subtype(relation):
    return len(specialisation_constraints(constraints(relation)))>0


def create_relation(name):
    return {RELATION_NAME: name, RELATION_ATTRIBUTES: [], RELATION_CONSTRAINTS: []}

def add_attribute(relation_type, name, data_type=ANY_TYPE):
    if RELATION_ATTRIBUTES not in relation_type:
        relation_type[RELATION_ATTRIBUTES] = []
    relation_type[RELATION_ATTRIBUTES].append({ATTRIBUTE_NAME: name, ATTRIBUTE_DATA_TYPE : data_type})

def add_primary_key_constraint(relation_type, attribute):
    relation_type[RELATION_CONSTRAINTS].append(create_primary_key_constraint(attribute))

def add_specialisation_constraint(relation_type, supertype_name):
    relation_type[RELATION_CONSTRAINTS].append(create_specialisation_constraint(supertype_name))

def add_coverage_constraint(relation_type, subtypes):
    relation_type[RELATION_CONSTRAINTS].append(create_coverage_constraint(subtypes))

def add_disjointness_constraint(relation_type, relation_type_names):
    assert(relation_type_name(relation_type) not in relation_type_names)
    relation_type[RELATION_CONSTRAINTS].append(create_disjointness_constraint(relation_type_names))


def add_foreign_key_constraint(relation_type, constraint_name, attribute_name, foreign_relation_type_name):
    relation_type[RELATION_CONSTRAINTS].append(create_foreign_key_constraint(constraint_name, attribute_name, foreign_relation_type_name))


def add_pathfd_constraint(relation_type, paths, implication_pathfd):
    # Could do all manner of checking that paths are valid....
    relation_type[RELATION_CONSTRAINTS].append(create_pathfd_constraint(paths, implication_pathfd))

def validate(attribute, model):
    pass


# Operations on constraints
FOREIGN_KEY = 'foreignKey'
ISA = 'isA'
COVERED_BY = 'coveredBy'
DISJOINT_WITH = 'disjointWith'
PATHFD = 'pathfd'
PRIMARY_KEY = 'primaryKey'
# --------------
CONSTRAINT_NAME = 'constraintName'
FOREIGN_KEY_ATTRIBUTE = 'attributeName'
FOREIGN_KEY_TARGET = 'relationName'

def constraint_to_string(constraint):
    if is_foreign_key(constraint):
        return foreign_key_string(constraint)
    elif is_coverage(constraint):
        return coverage_string(constraint)
    elif is_disjointness(constraint):
        return disjointness_string(constraint)
    elif is_specialisation(constraint):
        return specialisation_string(constraint)
    elif is_pathfd(constraint):
        return pathfd_string(constraint)
    return None

# Primary keys

def is_primary_key_constraint(constraint):
    return PRIMARY_KEY in constraint

def primary_key_attribute(constraint):
    assert(is_primary_key_constraint(constraint))
    return constraint[PRIMARY_KEY]

def primary_key_constraint(constraints):
    result = select_by_type( constraints, PRIMARY_KEY)
    assert(len(result)==1)
    return result[0]

def create_primary_key_constraint(attribute):
    return { PRIMARY_KEY: attribute}

def primary_key_string(constraint):
    assert(is_primary_key_constraint(constraint))
    return "primary key ({:s})".format(constraint[PRIMARY_KEY])

# Foreign keys


def foreign_key_constraints(constraints):
    return select_by_type(constraints, FOREIGN_KEY)

def is_foreign_key(constraint):
    return FOREIGN_KEY in constraint

def foreign_key_components(constraint):
    assert(FOREIGN_KEY in constraint)
    return constraint[FOREIGN_KEY][CONSTRAINT_NAME], constraint[FOREIGN_KEY][FOREIGN_KEY_ATTRIBUTE], constraint[FOREIGN_KEY][FOREIGN_KEY_TARGET]

def foreign_key_string(constraint):
    assert(is_foreign_key(constraint))
    name, fk_attribute, fk_relation_type = foreign_key_components(constraint)
    return "constraint {:s} foreign key {:s} references {:s}".format(name, fk_attribute, fk_relation_type)

def constraint_name(foreign_key_constraint):
    assert(is_foreign_key(foreign_key_constraint))
    return foreign_key_constraint[FOREIGN_KEY][CONSTRAINT_NAME]


def foreign_key_attribute(foreign_key_constraint):
    assert(is_foreign_key(foreign_key_constraint))
    return foreign_key_constraint[FOREIGN_KEY][FOREIGN_KEY_ATTRIBUTE]


def foreign_key_relation(foreign_key_constraint):
    assert(is_foreign_key(foreign_key_constraint))
    return foreign_key_constraint[FOREIGN_KEY][FOREIGN_KEY_TARGET]


def create_foreign_key_constraint(constraint_name, attribute_name, foreign_relation_type_name):
    result = { FOREIGN_KEY: { CONSTRAINT_NAME: constraint_name, FOREIGN_KEY_ATTRIBUTE: attribute_name, FOREIGN_KEY_TARGET: foreign_relation_type_name}}
    return result

def validate_foreign_key(foreign_key_constraint, relation, model):
    pass


# --------------
def specialisation_constraints(constraints):
    return select_by_type(constraints, ISA)


def is_specialisation(constraint):
    return ISA in constraint


def specialisation_string(constraint):
    assert(is_specialisation(constraint))
    return "isa {:s}".format(constraint[ISA])


def constraint_supertype(specialisation_constraint):
    assert(is_specialisation(specialisation_constraint))
    return specialisation_constraint[ISA]


def create_specialisation_constraint(supertype_name):
    return { ISA: supertype_name}

def validate_specialisation(specialisation_constraint, model):
    pass

# --------------
def coverage_constraints(constraints):
    return select_by_type(constraints, COVERED_BY)


def is_coverage(constraint):
    return COVERED_BY in constraint


def constraint_subtypes(coverage_constraint):
    assert(is_coverage(coverage_constraint))
    return list(coverage_constraint[COVERED_BY])

def coverage_string(constraint):
    assert(is_coverage(constraint))
    return 'covered by ({:s})'.format(join(constraint_subtypes(constraint), ', '))


def create_coverage_constraint(relation_type_names):
    relation_type_names = list(relation_type_names).copy()
    relation_type_names.sort()
    return { COVERED_BY : relation_type_names }

def validate_coverage_constraint(coverage_constraint, model):
    pass

# --------------
def disjointness_constraints(constraints):
    return select_by_type(constraints, DISJOINT_WITH)


def is_disjointness(constraint):
    return DISJOINT_WITH in constraint


def constraint_relations(disjointness_constraint):
    assert(is_disjointness(disjointness_constraint))
    return list(disjointness_constraint[DISJOINT_WITH])


def disjointness_string(constraint):
    assert(is_disjointness(constraint))
    return 'disjoint with ({:s})'.format(join(constraint_relations(constraint), ', '))


def create_disjointness_constraint(relation_type_names):
    relation_type_names = list(relation_type_names).copy()
    relation_type_names.sort()
    return { DISJOINT_WITH: list(relation_type_names.copy()) }

def validate_disjointness_constraint(disjointness_constraint, model):
    pass


# --------------
CONSTRAINT_PATHS = 'paths'
CONSTRAINT_IMPLICATION = 'imply'


def is_pathfd(constraint):
    return PATHFD in constraint


def pathfd_constraints(constraints):
    return select_by_type(constraints, PATHFD)


def constraint_paths(pathfd_constraint):
    assert(is_pathfd(pathfd_constraint))
    return pathfd_constraint[PATHFD][CONSTRAINT_PATHS]


def constraint_implication(pathfd_constraint):
    assert(is_pathfd(pathfd_constraint))
    return pathfd_constraint[PATHFD][CONSTRAINT_IMPLICATION]


def is_identifying(pathfd_constraint, primary_key_attribute):
    assert(is_pathfd(pathfd_constraint))
    implication_path = constraint_implication(pathfd_constraint)
    if not is_simple(implication_path) or attribute_name(implication_path) != primary_key_attribute:
        return False
    else:
        assert(constraint_paths(pathfd_constraint))
        result = True
        for path in constraint_paths(pathfd_constraint):
            if not is_simple(path) or attribute_name(path) == 'self':  # or not path in self.attributes:
                result = False
        return result

def pathfd_string(constraint):
    assert(is_pathfd(constraint))
    return 'pathfd ({:s}) -> {:s}'.format(
        path_list_string(constraint_paths(constraint)),
        path_string(constraint_implication(constraint)))


def create_pathfd_constraint(paths, implication):
    return { PATHFD: { CONSTRAINT_PATHS: paths.copy(), CONSTRAINT_IMPLICATION: implication}}


def add_constraint_path(pathfd_constraint, path):
    assert(is_pathfd(pathfd_constraint))
    constraint_paths(pathfd_constraint).append(path)


def validate_pathfd_constraint(pathfd_constraint, relation, model):
    pass

# --------------
def select_by_type(constraints, constraint_type):
    return [ constraint for constraint in constraints if constraint_type in constraint ]


# Operations on paths

def is_simple(path):
    assert(type(path)==list)
    return len(path)==1

def attribute_name(path):
    assert(is_simple(path))
    return path[0]


def create_simple_path(attribute_name):
    return [attribute_name]

def create_simple_paths(attribute_names):
    attribute_names = list(attribute_names)
    attribute_names.sort()
    simple_paths = list()
    for attribute_name in attribute_names:
        simple_paths.append([attribute_name])
    return simple_paths

def path_string(path):
    if len(path) > 0:
        result = path[0]
        for attribute_name in path[1:]:
            result += '.'+attribute_name
        return result
    else:
        return ''


def path_list_string(paths):
    result = ''
    if len(paths)>0:
        result = result+path_string(paths[0])
        for path in paths[1:]:
            result = result+', '+path_string(path)
    result = result+''
    return result


def join(strings, separator):
    if not strings:
        return ""
    else:
        iterator = strings.__iter__()
        result = next(iterator)
        while True:
            try:
                value = next(iterator)
                result = result+separator+value
            except StopIteration:
                break
        return result

