import unittest
import json
import ARM_To_EER

PATH = "../Samples/ARM_To_EER/"


class ARM_To_EER_Test(unittest.TestCase):
    '''Unit Tests for the ARM_to_EER class.'''


    @classmethod
    def readJSON(cls, filename):
        with open(filename, 'r') as f:
            datastore = json.load(f)
        return datastore


    def setUp(self):
        self.maxDiff = None


    """
    A. Single relation mapping to entity type    
    Input:
    relation Customer(self OID, customer_id string, name string, address string, postcode number,
        primary key (self),
        pathfd (customer_id)->self)
    Output:
    entity Customer(customer_id, name, address, postcode, identity(customer_id, complete))
    """
    def test_A(self):
        TEST_TAG = PATH+"ARM_Sample_A_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    B. Two relations, one with foreign key to other. Maps to two entity types and a relationship type
    1. 1:N binary relationship type
    
    
    Input
    relation Driver(self OID, name string, licence_id string, class number, za_id string,
        primary key (self),
        pathfd(za_id)->self),
        disjoint with (Mechanic))
    relation Mechanic(self OID, name string, class number, experience number, driver OID,
        primary key (self),
        pathfd(name, class)->self,
        constraint team foreign key driver references Driver,
        disjoint with (Driver))
    
    Output:
    entity Driver(name, licence_id, class, za_id, identity(za_id, complete))
    entity Mechanic(name, class, experience, identity(name, class, complete))
    relationship Team(participant(Driver, strong), participant(Mechanic, strong, 1, 1))
    """
    def test_B1(self):
        TEST_TAG = PATH+"ARM_Sample_B1_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)


    """
    B2. Three relations, one with foreign keys from others. Mapping to entity types and a relationship type. 
    
    Input:
    relation Driver(self OID, name string, licence_id string, class number, za_id string,
        primary key (self),
        pathfd(za_id)->self,
        disjoint with (Mechanic, Team))
    relation Mechanic(self OD, name string, class number, position string,
        primary key (self),
        pathfd(name, position)->self,
        disjoint with (Driver, Team))
    relation Team(self OID, driver OID, mechanic OID,
        primary key (self),
        pathfd(driver, mechanic)->self,
        constraint drives foreign key driver references Driver,
        constraint services foreign key mechanic references Mechanic,
        disjoint with (Driver, Mechanic))
    
    Output:
    entity Driver(name, licence_id, class, za_id, identity(za_id, complete))
    entity Mechanic(name, class, position, identity(name, position, complete))
    relationship Team(participant(Driver, strong, drives), participant(Mechanic, strong, services)
    """
    def test_B2(self):
        TEST_TAG = PATH+"ARM_Sample_B2_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    3. Two relations, one with foreign key to other and identity dependent on foreign key. Mapping to weak entity type in identifying binary relationship type.
    
    Input:
    relation Employee(self OID, name string, id string, salary number, 
        primary key (self),
        pathfd(id)->self,
        disjoint with (Dependent))
    relation Dependent(self OID, name string, dob date, employee OID,
        primary key (self),
        pathfd(name, employee)->self,
        constraint depencency foreign key employee references Employee,
        disjoint with (Employee))
    
    Output:
    entity Employee(name, id, salary, identity(id, complete))
    entity Dependent(name, dob, identity(name, partial))
    relationship Dependency(participant(Employee, strong), participant(Dependent, weak, 1, 1))
    """
    def test_B3(self):
        TEST_TAG = PATH+"ARM_Sample_B3_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    C. Three relations, one with foreign keys from others and identity dependent on them. Mapping to two entity types and a relationship type with attributes.
    
    Input:
    relation Driver(self OID, name string, licence_id string, class number, za_id string,
        primary key (self),
        pathfd(za_id)->self,
        disjoint with (Mechanic, Team))
    relation Mechanic(self OID, name string, class number, za_id string, 
        primary key (self),
        pathfd(za_id)->self,
        disjoint with (Driver, Team))
    relation Team(self OID, name string, rank number, driver OID, mechanic OID,
        primary key (self),
        pathfd(name)->self,
        constraint supports foreign key mechanic references Mechanic,
        constraint drives foreign key driver references Driver,
        disjoint with (Driver, Mechanic))
        
    Output:
    entity Driver(name, licence_id, class, za_id, identity(za_id, complete))
    entity Mechanic(name, class, za_id, identity(za_id, complete))
    relationship Team(name, rank, participant(Driver, drives, strong), participant(Mechanic, supports, strong))
    """
    def test_C(self):
        TEST_TAG = PATH+"ARM_Sample_C_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)


    """
    D1. Three relations, two with 'isa' constraints referencing the other. Mapping to 3 entity types and a specialisation.

    Input:
    relation Employee(self OID, emp_id anyType, name anyType, job_title anyType,
        primary key (self),
        pathfd(emp_id)->self)
    relation SalariedEmployee(self OID, salary anyType,
        primary key (self),
        isa Employee)
    relation CasualEmployee(self OID, hourly_rate anyType,
        primary key (self),
        isa Employee)
    
    Output:
    entity Employee(emp_id, name, job_title, identity(emp_id, complete))
    entity SalariedEmployee(salary)
    entity CasualEmployee(hourly_rate)
    specialisation (Employee, (SalariedEmployee, CasualEmployee))
    """
    def test_D1(self):
        TEST_TAG = PATH+"ARM_Sample_D1_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    D2. Four relations, three with 'isa' constraints referencing the other and disjoint with each other. Mapping to 4 entity types and a specialisation with disjointness.
    
    Input:
    relation Employee(self OID, emp_id anyType, name anyType, job_title anyType,
        primary key (self),
        pathfd(emp_id)->self)
    relation SalariedEmployee(self OID, salary anyType,
        primary key (self),
        isa Employee,
        disjoint with (ContractedEmployee, CommissionEmployee))
    relation ContractedEmployee(self OID, hourly_rate anyType,
        primary key (self),
        isa Employee,
        disjoint with (SalariedEmployee, CommissionEmployee))
    relation CommissionEmployee(self OID, percentage anyType,
        primary key (self),
        isa Employee,
        disjoint with (SalariedEmployee, ContractedEmployee))
    
    Output:
    entity Employee(emp_id, name, job_title, identity(emp_id, complete))
    entity SalariedEmployee(salary)
    entity ContractedEmployee(hourly_rate)
    entity CommissionEmployee(percentage)
    specialisation (Employee, (SalariedEmployee, ContractedEmployee, CommissionEmployee), disjoint)
    """

    def test_D2(self):
        TEST_TAG = PATH+"ARM_Sample_D2_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    D3. Three relations, one with coverage constraint referencing the other two, and the other two with 'isa' constraints referencing the first. Mapping to three entity types and a specialisation with complete coverage. 
    (Example based on figure 8.5 Fundamentals of Database Systems, 6th Ed., Elmasri, Navathe.)
    
    Input:
    relation Part(self OID, part_no anyType, description anyType, 
        primary key (self),
        pathd(part_no)->self,
        covered by (ManufacturedPart, PurchasedPart))
    relation ManufacturedPart(self OID, drawing_no anyType, batch_no anyType, manufacture_date anyType,
        primary key (self),
        isa Part)
    relation PurchasedPart(self OID, supplier_name, list_price,
        primary key (self),
        isa Part)
    
    Output:
    entity Part(part_no, description, identity(part_no, complete))
    entity ManufacturedPart(drawing_no, batch_no, manufacture_date)
    entity PurchasedPart(supplier_name, list_price)
    specialisation (Part, (ManufacturedPart, PurchasedPart), coverage(complete))
    """
    def test_D3(self):
        TEST_TAG = PATH+"ARM_Sample_D3_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)


    """
    E. A Bit of Everything
    
    Input:
    relation Employee(self OID, emp_id number, name string, job_title string, manager OID, department OID,
        primary key (self),
        pathfd(emp_id)->self,
        constraint management foreign key manager references Manager,
        constraint worksfor foreign key department references Department,
        covered by (SalariedEmployee, CasualEmployee),
        disjoint with (Department, Dependent, Project, WorksOn) 
    
    relation SalariedEmployee(self OID, salary number,
        primary key (self),
        isa Employee,
        disjoint with (CasualEmployee),
        disjoint with (Department, Dependent, Project, WorksOn))
    relation CasualEmployee(self OID, hourly_rate number,
        primary key (self),
        isa Employee,
        disjoint with (SalariedEmployee),
        disjoint with (Department, Dependent, Project, WorksOn))
    
    relation Manager(self OID, title anyType,
        primary key (self),
        isa Employee,
        disjoint with (Department, Dependent, Project, WorksOn))
    
    relation Dependent(self OID, name string, dob date, employee OID,
        primary key (self),
        pathfd(name, dob, employee)->self,
        constraint dependentsof foreign key employee references Employee,
        disjoint with (CasualEmployee, Department, Employee, Manager, Project, SalariedEmployee, WorksOn))
        
    relation Department(self OID, name string, number number, location string,
        primary key (self),
        pathfd(name, number)->self,
        disjoint with (CasualEmployee, Dependent, Employee, Manager, Project, SalariedEmployee, WorksOn))
    
    relation Project(self OID, name string, number number, department OID,
        primary key (self),
        pathfd(name, number)->self,
        constraint controls foreign key department references Department,
        disjoint with (CasualEmployee, Department, Dependent, Employee, Manager, SalariedEmployee, WorksOn))
    
    relation WorksOn(self OID, project OID, employee OID,
        primary key (self),
        pathfd(project, employee)->self,
        constraint deployment foreign key project references Project,
        constraint deployed foreign key employee references Employee,
        disjoint with (CasualEmployee, Department, Dependent, Employee, Manager, Project, SalariedEmployee))
    
    Output:
    entity Employee(emp_id, name, job_title, identity(emp_id, complete))
    entity SalariedEmployee(salary)
    entity CasualEmployee(hourly_rate)
    specialisation(Employee, {SalariedEmployee, CasualEmployee}, coverage(complete), disjointness(disjoint))
    
    entity Manager(title)
    specialisation(Employee, {Manager}, coverage(partial))
    relationship Management(participant(Manager,strong), participant(Employee, strong, 1, 1))
    
    entity Dependent(name, dob, identity(name, dob, partial))
    relationship DependentsOf(participant(Employee, provides_for, strong), participant(Dependent, supported, weak, 1, 1))
    
    entity Department(name, number, location, identity(name, number))
    entity Project(name, number, identity(name, number))
    
    relationship Controls(participant(Department, strong), participant(Project, strong, 1, 1))
    relationship WorksFor(participant(Employee, strong, 1, 1), participant(Department, strong))
    relationship WorksOn(participant(Project, deployment, strong), participant(Employee, deployed, strong))
    """
    def test_E(self):
        TEST_TAG = PATH+"ARM_Sample_E_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = ARM_To_EER.transform(arm_model)
        self.assertEqual(eer_model,transformation)

if __name__ == '__main__':
    unittest.main()



