import unittest
import json
import EER_To_ARM

PATH = "../Samples/EER_To_ARM/"


class EER_To_ARM_Test(unittest.TestCase):
    '''Unit Tests for the EER_to_ARM class.'''


    @classmethod
    def readJSON(cls, filename):
        with open(filename, 'r') as f:
            datastore = json.load(f)
        return datastore


    def setUp(self):
        self.maxDiff = None


    """
    A. Single entity type
    
    Input:
    entity Customer(customer_id, name, address, postcode, identity(customer_id, complete))
    
    Output:
    relation Customer(self OID, customer_id anyType, name anyType, address anyType, postcode anyType,
        primary key (self),
        pathfd (customer_id)->self)
    """
    def test_A(self):
        TEST_TAG = PATH+"EER_Sample_A_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)


    """
    B. Entity type in binary relationship type
    1. 1:N binary relationship type
    
    Input:
    entity Driver(name, licence_id, class, za_id, identity(za_id, complete))
    entity Mechanic(name, class, experience, identity(name, class, complete))
    relationship Team(participant(Driver, drives, strong, 0, N), participant(Mechanic, serves, strong, 1, 1))
    
    Output
    relation Driver(self OID, name anyType, licence_id anyType, class anyType, za_id anyType,
        primary key (self),
        pathfd(za_id)->self),
        disjoint with (Mechanic))
    relation Mechanic(self OID, name anyType, class anyType, experience anyType, driver OID,
        primary key (self),
        pathfd(name, class)->self,
        constraint team foreign key driver references Driver,
        disjoint with (Driver))
    """
    def test_B1(self):
        TEST_TAG = PATH+"EER_Sample_B1_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    B2. M:N binary relationship type
    
    Input:
    entity Driver(name, licence_id, class, za_id, identity(za_id, complete))
    entity Mechanic(name, class, position, identity(name, position, complete))
    relationship Team(participant(Driver, strong, 0, N), participant(Mechanic, strong 0, N)
    
    Output:
    relation Driver(self OID, name anyType, licence_id anyType, class anyType, za_id anyType,
        primary key (self),
        pathfd(za_id)->self,
        disjoint with (Mechanic, Team))
    relation Mechanic(self OD, name anyType, class anyType, position anyType,
        primary key (self),
        pathfd(name, class)->self,
        disjoint with (Driver, Team))
    relation Team(self OID, driver OID, mechanic OID,
        primary key (self),
        pathfd(driver, mechanic)->self,
        disjoint with (Driver, Mechanic))
    
    """
    def test_B2(self):
        TEST_TAG = PATH+"EER_Sample_B2_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    B3. Weak entity type in identifying binary relationship type
    
    Input:
    entity Employee(name, id, salary, identity(id, complete))
    entity Dependent(name, dob, identity(name, partial))
    relationship Dependency(participant(Employee, strong, 0, N), participant(Dependent, weak, 1, 1))
    
    Output:
    relation Employee(self OID, name anyType, id anyType, salary anyType, 
        primary key (self),
        pathfd(id)->self,
        disjoint with (Dependent))
    relation Dependent(self OID, name anyType, dob anyType, employee OID,
        primary key (self),
        pathfd(name, employee)->self,
        constraint depencency foreign key employee references Employee,
        disjoint with (Employee))
    """
    def test_B3(self):
        TEST_TAG = PATH+"EER_Sample_B3_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    C. Entity type in 1:N binary relationship type which itself has attributes
    
    Input:
    entity Driver(name, licence_id, class, za_id, identity(za_id, complete))
    entity Mechanic(name, class, za_id, identity(za_id, complete))
    relationship Team(name, rank, participant(Driver, strong, 0, N), participant(Mechanic, serves, strong, 1, 1))
    
    Output:
    relation Driver(self OID, name anyType, licence_id anyType, class anyType, za_id anyType,
        primary key (self),
        pathfd(za_id)->self,
        disjoint with (Mechanic))
    relation Mechanic(self OID, name anyType, class anyType, za_id anyType, 
        team_name anyType, team_rank anyType, driver OID, 
        primary key (self),
        pathfd(za_id)->self,
        constraint team foreign key driver references Driver,
        disjoint with (Driver))
    """
    def test_C(self):
        TEST_TAG = PATH+"EER_Sample_C_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    D. Subtypes
    1. 2 subtypes overlapping and partial coverage
    
    
    Input:
    entity Employee(emp_id, name, job_title, identity(emp_id, complete))
    entity SalariedEmployee(salary)
    entity CasualEmployee(hourly_rate)
    specialisation (Employee, (SalariedEmployee, CasualEmployee))
    
    Output:
    relation Employee(self OID, emp_id anyType, name anyType, job_title anyType,
        primary key (self),
        pathfd(emp_id)->self)
    relation SalariedEmployee(self OID, salary anyType,
        primary key (self),
        isa Employee)
    relation CasualEmployee(self OID, hourly_rate anyType,
        primary key (self),
        isa Employee)
    """
    def test_D1(self):
        TEST_TAG = PATH+"EER_Sample_D1_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    D2. 3 disjoint subtypes
    
    Input:
    entity Employee(emp_id, name, job_title, identity(emp_id, complete))
    entity SalariedEmployee(salary)
    entity CasualEmployee(hourly_rate)
    entity CommissionEmployee(percentage)
    specialisation (Employee, (SalariedEmployee, CasualEmployee, CommissionEmployee), disjoint)
    
    Output:
    relation Employee(self OID, emp_id anyType, name anyType, job_title anyType,
        primary key (self),
        pathfd(emp_id)->self)
    relation SalariedEmployee(self OID, salary anyType,
        primary key (self),
        isa Employee,
        disjoint with (CasualEmployee, CommissionEmployee))
    relation CasualEmployee(self OID, hourly_rate anyType,
        primary key (self),
        isa Employee,
        disjoint with (SalariedEmployee, CommissionEmployee))
    relation CommissionEmployee(self OID, percentage anyType,
        primary key (self),
        isa Employee,
        disjoint with (SalariedEmployee, CasualEmployee))
    """
    def test_D2(self):
        TEST_TAG = PATH+"EER_Sample_D2_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    D3. 2 overlapping subtypes with complete coverage 
    (Example based on figure 8.5 Fundamentals of Database Systems, 6th Ed., Elmasri, Navathe.)
    
    Input:
    
    entity Part(part_no, description, identity(part_no, complete))
    entity ManufacturedPart(drawing_no, batch_no, manufacture_date)
    entity PurchasedPart(supplier_name, list_price)
    specialisation (Part, (ManufacturedPart, PurchasedPart), coverage(complete), disjointness(overlapping))
    
    Output:
    relation Part(self OID, part_no anyType, description anyType, 
        primary key (self),
        pathd(part_no)->self,
        covered by (ManufacturedPart, PurchasedPart))
    relation ManufacturedPart(self OID, drawing_no anyType, batch_no anyType, manufacture_date anyType,
        primary key (self),
        isa Part)
    relation PurchasedPart(self OID, supplier_name, list_price,
        primary key (self),
        isa Part)
    """
    def test_D3(self):
        TEST_TAG = PATH+"EER_Sample_D3_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)

    """
    E. A Bit of Everything
    
    Input:
    entity Employee(emp_id, name, job_title, identity(emp_id, complete))
    entity SalariedEmployee(salary)
    entity CasualEmployee(hourly_rate)
    specialisation(Employee, {SalariedEmployee, CasualEmployee}, coverage(complete), disjointness(disjoint))
    
    entity Manager(title)
    specialisation(Employee, {Manager}, coverage(partial))
    relationship Management(participant(Manager,strong, 1, N), participant(Employee, strong, 1, 1))
    
    entity Dependent(name, dob, identity(name, dob, partial))
    relationship DependentsOf(date_added, participant(Employee, provides_for, strong, 0, N), participant(Dependent, supported, weak, 1, 1))
    
    entity Department(name, number, location, identity(name, number))
    entity Project(name, number, identity(name, number))
    
    relationship Controls(participant(Department, 1, N), participant(Project, 1, 1))
    relationship WorksFor(participant(Employee, 1, 1), participant(Department, 1, N))
    relationship WorksOn(participant(Project, 1, N), participant(Employee, 0, N))
    
    Output:
    relation Employee(self OID, emp_id anyType, name anyType, job_title anyType, manager OID, department OID,
        primary key (self),
        pathfd(emp_id)->self,
        constraint management foreign key manager references Manager,
        constraint worksfor foreign key department references Department,
        disjoint with (Department, Dependent, Project, WorksOn) 
    
    relation SalariedEmployee(self OID, salary anyType,
        primary key (self),
        isa Employee,
        disjoint with (Department, Dependent, Project, WorksOn))
    relation CasualEmployee(self OID, hourly_rate anyType,
        primary key (self),
        isa Employee,
        disjoint with (Department, Dependent, Project, WorksOn))
    
    relation Manager(self OID, title anyType,
        primary key (self),
        isa Employee,
        disjoint with (Department, Dependent, Project, WorksOn))
    
    relation Dependent(self OID, name anyType, dob anyType, employee OID, dependentsof_date_added anyType,
        primary key (self),
        pathfd(name, employee)->self,
        constraint depencency foreign key employee references Employee,
        disjoint with (Department, Employee, SalariedEmployee, CasualEmployee, Manager, Project, WorksOn))
        
    relation Department(self OID, name anyType, number anyType, location anyType,
        primary key (self),
        pathfd(name, number)->self,
        disjoint with (Dependent, Employee, SalariedEmployee, CasualEmployee, Manager, Project, WorksOn))
    
    relation Project(self OID, name anyType, number anyType, department OID,
        primary key (self),
        pathfd(name, number)->self,
        constraint controls foreign key department references Department,
        disjoint with (Department, Dependent, Employee, SalariedEmployee, CasualEmployee, Manager, WorksOn))
    
    relation WorksOn(self OID, project OID, employee OID,
        primary key (self),
        pathfd(project, employee)->self,
        disjoint with (Department, Dependent, Employee, SalariedEmployee, CasualEmployee, Manager, Project, WorksOn))
    """
    def test_E(self):
        TEST_TAG = PATH+"EER_Sample_E_{:s}.JSON"
        arm_model = self.readJSON(TEST_TAG.format('Input'))
        eer_model = self.readJSON(TEST_TAG.format('Output'))
        transformation = EER_To_ARM.transform(arm_model)
        self.assertEqual(eer_model,transformation)
